package tofly.opc_gateway;

import org.jinterop.dcom.common.JIException;
import org.jinterop.dcom.core.JIVariant;
import org.openscada.opc.lib.common.AlreadyConnectedException;
import org.openscada.opc.lib.common.ConnectionInformation;
import org.openscada.opc.lib.common.NotConnectedException;
import org.openscada.opc.lib.da.*;
import org.springframework.boot.ApplicationRunner;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.context.annotation.Bean;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import tofly.opc_gateway.config.OpcServerProperties;
import tofly.opc_gateway.log.Log;
import tofly.opc_gateway.log.LogFactory;

import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;

@SpringBootApplication
@EnableScheduling
public class Application {

    private static final Log log = LogFactory.getLog(Application.class);

    private static final String SPRING_CONFIG_NAME_KEY = "--spring.config.name";
    private static final String DEFAULT_SPRING_CONFIG_PARAM = SPRING_CONFIG_NAME_KEY + "=" + "opc_gateway";

    public static void main(String[] args) {
        SpringApplication.run(Application.class, updateArguments(args));
    }


    @Bean
    public ApplicationRunner applicationRunner(OpcServerProperties opcServerProperties) {
        return args -> {
            opcServerProperties.getMappings().forEach((k, v) -> {
                Thread thread = new Thread(() -> {
                    Server server = createServer(v.getHost(), v.getUser(), v.getPassword(), v.getClsid());
                    try {
                        server.connect();
                        final AccessBase access = new SyncAccess(server, v.getPeriod());
                        Map<String, List<Map<String, String>>> deviceMappings = v.getDeviceMappings();
                        deviceMappings.forEach((k1, v1) -> {
                            Map<String, String> itemMappings = new LinkedHashMap<>();
                            DataCallback dataCallback = new ServerDataCallback(k1,itemMappings);
                            v1.forEach(f -> {
                                f.forEach((k2, v2) -> {
                                    itemMappings.put(v2, k2);
                                    try {
                                        access.addItem(v2, dataCallback);
                                    } catch (JIException e) {
                                        throw new RuntimeException(e);
                                    } catch (AddFailedException e) {
                                        throw new RuntimeException(e);
                                    }
                                });
                            });
                        });
                        access.bind();
                        while (true) {
                            Thread.yield();
                        }

                    } catch (UnknownHostException e) {
                        throw new RuntimeException(e);
                    } catch (JIException e) {
                        throw new RuntimeException(e);
                    } catch (AlreadyConnectedException | NotConnectedException | DuplicateGroupException e) {
                        throw new RuntimeException(e);
                    }
                }, "OPC-Server[" + k + "]");

                thread.start();

            });

        };
    }

    @Bean(name = "threadPoolTaskScheduler")
    public ThreadPoolTaskScheduler threadPoolTaskScheduler() {
        ThreadPoolTaskScheduler threadPoolTaskScheduler = new ThreadPoolTaskScheduler();
        threadPoolTaskScheduler.setPoolSize(Runtime.getRuntime().availableProcessors());
        threadPoolTaskScheduler.setThreadNamePrefix("WorKerThread:");
        threadPoolTaskScheduler.setWaitForTasksToCompleteOnShutdown(true);
        threadPoolTaskScheduler.setAwaitTerminationSeconds(30);
        return threadPoolTaskScheduler;
    }

    private Server createServer(String host, String user, String password, String clsid) {
        final ConnectionInformation ci = new ConnectionInformation();
        ci.setHost(host);            // 电脑IP
        ci.setDomain("");                   // 域，为空就行
        ci.setUser(user);              // 用户名，配置DCOM时配置的
        ci.setPassword(password);           // 密码
        ci.setClsid(clsid); // KEPServer的注册表ID，可以在“组件服务”里看到
        return new Server(ci, Executors.newSingleThreadScheduledExecutor());
    }

    private static String[] updateArguments(String[] args) {
        if (Arrays.stream(args).noneMatch(arg -> arg.startsWith(SPRING_CONFIG_NAME_KEY))) {
            String[] modifiedArgs = new String[args.length + 1];
            System.arraycopy(args, 0, modifiedArgs, 0, args.length);
            modifiedArgs[args.length] = DEFAULT_SPRING_CONFIG_PARAM;
            return modifiedArgs;
        }
        return args;
    }
}