package tofly;

import com.google.common.collect.ImmutableList;
import org.eclipse.milo.opcua.sdk.client.OpcUaClient;
import org.eclipse.milo.opcua.sdk.client.api.identity.AnonymousProvider;
import org.eclipse.milo.opcua.sdk.client.nodes.UaNode;
import org.eclipse.milo.opcua.sdk.client.nodes.UaVariableNode;
import org.eclipse.milo.opcua.stack.core.Identifiers;
import org.eclipse.milo.opcua.stack.core.UaException;
import org.eclipse.milo.opcua.stack.core.security.SecurityPolicy;
import org.eclipse.milo.opcua.stack.core.types.builtin.DataValue;
import org.eclipse.milo.opcua.stack.core.types.builtin.NodeId;
import org.eclipse.milo.opcua.stack.core.types.enumerated.ServerState;
import org.eclipse.milo.opcua.stack.core.types.enumerated.TimestampsToReturn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.CompletableFuture;

import static org.eclipse.milo.opcua.stack.core.types.builtin.unsigned.Unsigned.uint;

public class ReadTest {
    private static final Logger logger = LoggerFactory.getLogger(ReadTest.class);

    public static void main(String[] args) throws Exception {
        OpcUaClient client = createClient();

        // synchronous connect
        client.connect().get();
        // 同步
        UaVariableNode node = client.getAddressSpace().getVariableNode(Identifiers.Server_ServerStatus_StartTime);
        DataValue value = node.readValue();
        logger.info("StartTime={}", value.getValue().getValue());

        //异步
        List<NodeId> nodeIds = ImmutableList.of(
                new NodeId(2, "通道 1.设备 1.标记 1"),
                new NodeId(2, "通道 1.设备 1.标记 3"));
        client.readValues(5000, TimestampsToReturn.Both, nodeIds).thenAccept(values -> {
                    DataValue v0 = values.get(0);
                    DataValue v1 = values.get(1);
                    logger.info("通道 1.设备 1.标记 1={}", v0.getValue().getValue());
                    logger.info("通道 1.设备 1.标记 2={}", v1.getValue().getValue());
                }).exceptionally(ex -> {
                    ex.printStackTrace();
                    return null;
                })
                .whenComplete((result, ex) -> {
                    if (ex != null) {
                        ex.printStackTrace();
                    }
                });
        Thread.sleep(10000);
        browseNode("", client, Identifiers.RootFolder);
    }

    private static OpcUaClient createClient() throws Exception {
        return OpcUaClient.create(
                "opc.tcp://127.0.0.1:49320/milo",
                endpoints ->
                        endpoints.stream()
                                .filter(e -> SecurityPolicy.None.getUri().equals(e.getSecurityPolicyUri()))
                                .findFirst(),
                configBuilder ->
                        configBuilder
                                .setIdentityProvider(new AnonymousProvider())
                                .setRequestTimeout(uint(5000))
                                .build()
        );
    }

    private static CompletableFuture<List<DataValue>> readServerStateAndTime(OpcUaClient client) {
        List<NodeId> nodeIds = ImmutableList.of(
                Identifiers.Server_ServerStatus_State,
                Identifiers.Server_ServerStatus_CurrentTime);

        return client.readValues(0.0, TimestampsToReturn.Both, nodeIds);
    }

    private static CompletableFuture<List<DataValue>> batchRead(OpcUaClient client) {
        List<NodeId> nodeIds = ImmutableList.of(
                new NodeId(2, "通道 1.设备 1.标记 1"));

        return client.readValues(1000, TimestampsToReturn.Both, nodeIds);
    }

    private static void browseNode(String indent, OpcUaClient client, NodeId browseRoot) {
        try {
            List<? extends UaNode> nodes = client.getAddressSpace().browseNodes(browseRoot);

            for (UaNode node : nodes) {
                logger.info("{} Node={}", indent, node.getNodeId());
                // recursively browse to children
                browseNode(indent + "  ", client, node.getNodeId());
            }
        } catch (UaException e) {
            logger.error("Browsing nodeId={} failed: {}", browseRoot, e.getMessage(), e);
        }
    }


}